Unit CIA;

Interface

{$I CBMOPT.PAS}

Procedure Reset;

Procedure EmulateLine(Cycles:word);

Function PeekCIA1(Addr:byte):byte;
Function PeekCIA2(Addr:word):byte;
Procedure PokeCIA1(Addr:word; Value:byte);
Procedure PokeCIA2(Addr:word; Value:byte);

Var
 CIAIRQ:Boolean;
 CIANMI:boolean;
 KeyMatrix:array[0..7] of byte;
 RevMatrix:array[0..7] of byte;

Implementation

uses
 memory,keyboard;

Type
 TCIA=Record
 case boolean of
  True:(
  DPA:byte; { Data Port A }
  DPB:byte; { Data Port B }
  DDA:byte; { Data Direction Register - Port A }
  DDB:byte; { Data Direction Register - Port B }
  TIA:word; { Timer A }
  TIB:word; { Timer B }
  TDD:byte; { Time of Day Clock : 1/10 sec }
  TDS:byte; { Time of Day Clock : Seconds  }
  TDM:byte; { Time of Day Clock : Minutes  }
  TDH:byte; { Time of Day Clock : Hours    }
  SSD:byte; { Synchronous Control Register }
  ICR:byte; { Interrupt Control Register }
  CRA:byte; { Control Register A }
  CRB:byte; { Control Register B }
 { Timers }
  iMask :byte;
  TimerA:word;
  TimerB:word;
  StartA:boolean;
  StartB:boolean;
  Underflow:boolean;
 { Alarm }
  ALD:byte;
  ALS:byte;
  ALM:byte;
  ALH:byte;
  );
  False:(
   raw:array[0..15] of byte;
  );
 end;

Var
 CIA1:TCIA; { $DC00..$DC0E }
 CIA2:TCIA; { $DD00..$DD0E }

 Joystick1:byte;
 Joystick2:byte;

Procedure Reset;
 begin
  CIAIRQ:=False;
  CIANMI:=False;
  FillChar(CIA1,SizeOf(CIA1),0);
  FillChar(CIA2,SizeOf(CIA2),0);
  Joystick1:=$FF;
  Joystick2:=$FF;
  FillChar(KeyMatrix,Sizeof(RevMatrix),$FF);
  FillChar(RevMatrix,Sizeof(RevMatrix),$FF);
 end;

Function PeekCIA1(Addr:byte):byte;
 Var
  Result:byte;
  Test  :byte;
  i     :byte;
 begin
  { $DC00..$DCFF }
  with CIA1 do
  case Addr and $F of
   $00: begin
         Result:=DPA or (not DDA);
         Test  :=DPB or (not DDB){ and Joystick1};
         if (Test and $01)=0 then Result:=Result and (RevMatrix[0]);
         if (Test and $02)=0 then Result:=Result and (RevMatrix[1]);
         if (Test and $04)=0 then Result:=Result and (RevMatrix[2]);
         if (Test and $08)=0 then Result:=Result and (RevMatrix[3]);
         if (Test and $10)=0 then Result:=Result and (RevMatrix[4]);
         if (Test and $20)=0 then Result:=Result and (RevMatrix[5]);
         if (Test and $40)=0 then Result:=Result and (RevMatrix[6]);
         if (Test and $80)=0 then Result:=Result and (RevMatrix[7]);
         {Result:=Result and Joystick2;}
        end;
   $01: begin
         Result:=(not DDB);
         Test  :=DPA or (not DDA){ and Joystick2};
         if (Test and $01)=0 then Result:=Result and (KeyMatrix[0]);
         if (Test and $02)=0 then Result:=Result and (KeyMatrix[1]);
         if (Test and $04)=0 then Result:=Result and (KeyMatrix[2]);
         if (Test and $08)=0 then Result:=Result and (KeyMatrix[3]);
         if (Test and $10)=0 then Result:=Result and (KeyMatrix[4]);
         if (Test and $20)=0 then Result:=Result and (KeyMatrix[5]);
         if (Test and $40)=0 then Result:=Result and (KeyMatrix[6]);
         if (Test and $80)=0 then Result:=Result and (KeyMatrix[7]);
         Result:=(Result or (DPB and DDB)){ and Joystick2};
        end;

  (*
   $02: Result:=DDA;
   $03: Result:=DDB;
   $04: Result:=TimerA and $FF;
   $05: Result:=TimerA shr 8;
   $06: Result:=TimerB and $FF;
   $07: Result:=TimerB shr 8;
   $08: begin { tod_halt:=false; } Result:=TDD; end;
   $09: Result:=TDS;
   $0A: Result:=TDM;
   $0B: begin { tod_halt:=true; } Result:=TDH; end;
   $0C: Result:=SSD;
  *)
   $0D: begin
         Result:=ICR;
         ICR:=0;
         CIAIRQ:=False;
        end;
  (*
   $0E: Result:=CRA;
   $0F: Result:=CRB;
  *)
   else Result:=Raw[Addr and $F];
  end;
  PeekCIA1:=Result;
 end;

Function PeekCIA2(Addr:word):byte;
 var
  result:byte;
 begin
  with CIA2 do
   Case Addr and $F of
    $00: Result:=DPA or (not DDA) and $3F;{| IECLines & the_cpu_1541->IECLines;}
    $01: Result:=DPB or (not DDB);
   (*
    $02: Result:=DDA;
    $03: Result:=DDB;
    $04: Result:=TimerA and $FF;
    $05: Result:=TimerA shr 8;
    $06: Result:=TimerB and $FF;
    $07: Result:=TimerB shr 8;
    $08: begin { tod_halt:=false }Result:=TDD; end;
    $09: Result:=TDS;
    $0A: Result:=TDM;
    $0B: begin { tod_halt:=true } Result:=TDH; end;
    $0C: Result:=SSD;
   *)
    $0D: begin
          Result:=ICR;
          ICR:=0;
          CIANMI:=False;
         end;
   (*
    $0E: Result:=CRA;
    $0F: Result:=CRB;
   *)
    else Result:=Raw[Addr and $F];
  end;
  PeekCIA2:=Result;
 end;

Procedure PokeCIA1(Addr:word; Value:byte);
 begin
  with CIA1 do
  case Addr of
   $00 : DPA:=Value;
   $01 : begin DPB:=Value; { CheckLP; } end;
   $02 : DDA:=Value;
   $03 : begin CIA1.DDB:=Value; { CheckL; } end;
   $04 : TIA:=(TIA and $FF00) or Value;
   $05 : begin
          TIA:=(TIA and $FF) or (word(Value) shl 8);
          if (StartA=False) then TimerA:=TIA; { Reload timer if stopped }
         end;
   $06 : TIB:=(TIB and $FF00) or Value;
   $07 : begin
          TIB:=(TIB and $FF) or (word(Value) shl 8);
          if (StartB=False) then TimerB:=TIB; { Reload timer if stopped }
         end;
   $08 : if (CRB and $80)=0 then TDD:=Value and $0F else ALD:=Value and $0F;
   $09 : if (CRB and $80)=0 then TDS:=Value and $0F else ALS:=Value and $0F;
   $0A : if (CRB and $80)=0 then TDM:=Value and $0F else ALM:=Value and $0F;
   $0B : if (CRB and $80)=0 then TDH:=Value and $0F else ALH:=Value and $0F;
   $0C : begin CIA1.SSD:=Value; { TriggerInterrupt(8) } end;
   $0D : if (Value and $80)=0 then
          iMask:=iMask and (not value)
         else begin
          iMask:=iMask or (Value and $7F);
          if (ICR and iMask and $1F)<>0 then begin
           ICR:=ICR and $80;
           CIAIRQ:=True;
          end;
         end;
  $0E : begin
         CRA:=Value and $EF;
         if (Value and $10)=$10 then TimerA:=TIA; { Force Load }
         StartA:=(Value and $21)=1;
        end;
  $0F : begin
         CRB:=Value and $EF;
         if (Value and $10)=$10 then TimerB:=TIB; { Force Load }
         StartB:=(Value and $61)=1;
         Underflow:=(Value and $61)=$41;
        end;
  end;
 end;

Procedure PokeCIA2(Addr:word; Value:byte);
 begin
  with CIA2 do
  case Addr of
   $00: begin
         DPA:=Value;
         (*
			byte = ~pra & ddra;
			the_vic->ChangedVA(byte & 3);
			uint8 old_lines = IECLines;
			IECLines = (byte << 2) & 0x80	// DATA
				| (byte << 2) & 0x40		// CLK
				| (byte << 1) & 0x10;		// ATN
			if ((IECLines ^ old_lines) & 0x10) {	// ATN changed
				the_cpu_1541->NewATNState();
				if (old_lines & 0x10)				// ATN 1->0
					the_cpu_1541->IECInterrupt();
         *)
        end;
   $01: DPB:=Value;
   $02: begin DDA:=Value; { the_vic->ChangedVA(~(pra | ~ddra) & 3); } end;
   $03: DDB:=Value;
   $04 : TIA:=(TIA and $FF00) or Value;
   $05 : begin
          TIA:=(TIA and $FF) or (word(Value) shl 8);
          if (StartA=False) then TimerA:=TIA; { Reload timer if stopped }
         end;
   $06 : TIB:=(TIB and $FF00) or Value;
   $07 : begin
          TIB:=(TIB and $FF) or (word(Value) shl 8);
          if (StartB=False) then TimerB:=TIB; { Reload timer if stopped }
         end;
   $08 : if (CRB and $80)=0 then TDD:=Value and $0F else ALD:=Value and $0F;
   $09 : if (CRB and $80)=0 then TDS:=Value and $0F else ALS:=Value and $0F;
   $0A : if (CRB and $80)=0 then TDM:=Value and $0F else ALM:=Value and $0F;
   $0B : if (CRB and $80)=0 then TDH:=Value and $0F else ALH:=Value and $0F;
   $0C : begin CIA1.SSD:=Value; { TriggerInterrupt(8) } end;
   $0D : if (Value and $80)=0 then
          iMask:=iMask and (not value)
         else begin
          iMask:=iMask or (Value and $7F);
          if (ICR and iMask and $1F)<>0 then begin
           ICR:=ICR and $80;
           CIANMI:=True;
          end;
         end;
  $0E : begin
         CRA:=Value and $EF;
         if (Value and $10)=$10 then TimerA:=TIA; { Force Load }
         StartA:=(Value and $21)=1;
        end;
  $0F : begin
         CRB:=Value and $EF;
         if (Value and $10)=$10 then TimerB:=TIB; { Force Load }
         StartB:=(Value and $61)=1;
         Underflow:=(Value and $61)=$41;
        end;
  end;
 end;

Procedure EmulateLine(Cycles:word);
 Var
  UnderflowB:boolean;
 begin
 with CIA1 do begin { CIA1 }

  { Timer A }

  UnderflowB:=False;

  if StartA then begin

   if TimerA>Cycles then

    Dec(TimerA,Cycles)

   else begin

    TimerA:=TIA; { Reload Timer }

    if (CRA and 8)=8 then begin { One shot }
     CRA:=CRA and $FE;
     StartA:=False;
    end;

    { TriggerInterrupt(1) }
    ICR:=ICR or 1;
    if (iMask and 1)=1 then begin
     ICR:=ICR or $80;
     CIAIRQ:=True;{ CPU.TriggerCIAIRQ }
    end;

    if Underflow then
     if TimerB=0 then UnderflowB:=True else dec(TimerB);

   end;

  end;

  { Timer B }
  if StartB and (not UnderflowB) then
   if TimerB>Cycles then Dec(TimerB,Cycles) else UnderflowB:=True;

  { Timer B IRQ }
  if UnderflowB then begin

   TimerB:=TIB;

   if (CRB and 8)=8 then begin { One shot }
    CRB:=CRB and $FE;
    StartB:=False;
    Underflow:=False;
   end;

   { TriggerInterrupt(2) }
    ICR:=ICR or 2;
    if (iMask and 2)=2 then begin
     ICR:=ICR or $80;
     CIAIRQ:=True;{ CPU.TriggerCIAIRQ }
    end;

  end;

 end; { with CIA1 }

 with CIA2 do begin { CIA1 }

  { Timer A }

  UnderflowB:=False;

  if StartA then begin

   if TimerA>Cycles then

    Dec(TimerA,Cycles)

   else begin

    TimerA:=TIA; { Reload Timer }

    if (CRA and 8)=8 then begin { One shot }
     CRA:=CRA and $FE;
     StartA:=False;
    end;

    { TriggerInterrupt(1) }
    ICR:=ICR or 1;
    if (iMask and 1)=1 then begin
     ICR:=ICR or $80;
     CIANMI:=True;
    end;

    if Underflow then
     if TimerB=0 then UnderflowB:=True else dec(TimerB);

   end;

  end;

  { Timer B }
  if StartB and (not UnderflowB) then
   if TimerB>Cycles then Dec(TimerB,Cycles) else UnderflowB:=True;

  { Timer B IRQ }
  if UnderflowB then begin

   TimerB:=TIB;

   if (CRB and 8)=8 then begin { One shot }
    CRB:=CRB and $FE;
    StartB:=False;
    Underflow:=False;
   end;

   { TriggerInterrupt(2) }
    ICR:=ICR or 2;
    if (iMask and 2)=2 then begin
     ICR:=ICR or $80;
     CIANMI:=True;{ CPU.TriggerCIAIRQ }
    end;

  end;

 end; { with CIA2 }

 end;

end.