Unit CPU;

Interface

{$I CBMOPT}


uses
 memory,tools,haltcode,iec;

Procedure Run;
Procedure RunCase; { BP doesn't build a jump table :( }
Procedure PushW(w:word);

var
 Cycles:integer;

Implementation

Procedure SetFlags(AFlags:Flags; on:boolean);
 begin
  if on then P:=P+AFlags else P:=P-AFlags;
 end;

Function SetNZ(b:ShortInt):byte;
 begin
  SetFlags([NFg],b<0);
  SetFlags([ZFg],b=0);
  SetNZ:=b;
 end;

{ Memory access }

Function Imm :word; begin Imm :=Peek(PC); inc(pc); end;
(*
Function ZPge:word; begin ZPge:=Imm;   end;
Function ZPgX:word; begin ZPgX:=Imm+X; end;
Function ZPgY:word; begin ZPgY:=Imm+Y; end;
Function Abso:word; begin Abso:=Imm shl 8 +Imm; end;
Function AbsX:word; begin AbsX:=Abso+X; end;
Function AbsY:word; begin AbsY:=Abso+Y; end;
Function Ind :word; begin Ind :=PeekW(Abso);  end;
Function IndY:word; begin IndY:=PeekW(Imm)+Y; end;
Function IndX:word; begin IndX:=PeekW(Imm+X); end;
*)

Function ZPge:word; begin ZPge:=Peek(PC);   inc(pc) end;
Function ZPgX:word; begin ZPgX:=Peek(PC)+X; inc(pc) end;
Function ZPgY:word; begin ZPgY:=Peek(PC)+Y; inc(pc) end;
Function Abso:word; begin Abso:=Imm shl 8 +Imm; end;
Function AbsX:word; begin AbsX:=Imm shl 8 +Imm+X; end;
Function AbsY:word; begin AbsY:=Imm shl 8 +Imm+Y; end;
Function Ind :word; begin Ind :=PeekW(Imm shl 8 +Imm);  end;
Function IndY:word; begin IndY:=PeekW(peek(pc))+Y; inc(pc) end;
Function IndX:word; begin IndX:=PeekW(peek(pc)+X); inc(pc) end;

{ Read a byte in memory }

Function PeekZPge:byte; begin PeekZPge:=Bank0^[Peek(PC)]; inc(pc) end;
Function PeekNZZPge:byte; begin PeekNZZPge:=SetNZ(Bank0^[Peek(PC)]); inc(pc) end;

Function PeekNZ(at:word):byte; begin PeekNZ:=SetNZ(Peek(At)); end;
Procedure PushW(w:word); begin push(hi(w)); push(lo(w)); end;

{ Instructions }

Procedure BRK;
 begin
  pushw(PC+2);
  push(Byte(P) or Byte(BFg));
  PC:=PeekW($FFFE);
 end;

Procedure ADC(Value:byte);
 var
  som:word;
 begin
  som:=A+Value; if CFg in P then inc(som);
  SetFlags([CFg],(Som>255));
  SetFlags([VFg],(A and 128)<>(Som and 128));
  A:=SetNZ(som);
 end;

Procedure SBC(Value:byte);
 var
  sub:Integer;
 begin
  sub:=A-Value; if not (CFg in P) then dec(sub);
  SetFlags([CFg],(Sub>=0));
  SetFlags([VFg],(A and 128)<>(Sub and 128));
  A:=SetNZ(sub);
 end;

Procedure CMP(From,Value:byte);
 begin
  SetFlags([NFg],Value>From);
  SetFlags([ZFg],From=Value);
  SetFlags([CFg],From>=Value);
 end;

Procedure Branch(Cond:boolean);
 var rela:shortint;
 begin
  rela:=ShortInt(Imm);
  if Cond then inc(PC,rela);
 end;

Function ROL(b:ShortInt):byte;
 Var
  carry:boolean;
 begin
  carry:=CFg in P;
  SetFlags([CFg],b<0);
  b:=b SHL 1;
  if carry then b:=b or 1;
  ROL:=SetNZ(b);
 end;

Function ROR(b:byte):byte;
 var
  carry:boolean;
 begin
  carry:=CFg in P;
  SetFlags([CFg],odd(b));
  b:=b SHR 1;
  if carry then b:=b or 128;
  ROR:=SetNZ(b);
 end;

Function ASL(b:ShortInt):byte;
 begin
  SetFlags([CFg],b<0);
  ASL:=SetNZ(B SHL 1);
 end;

Function LSR(b:byte):byte;
 begin
  SetFlags([CFg],odd(b));
  LSR:=SetNZ(B SHR 1);
 end;

Procedure ROL_at(Addr:word); begin Poke(Addr,ROL(Peek(Addr))); end;
Procedure ROR_at(Addr:word); begin Poke(Addr,ROR(Peek(Addr))); end;
Procedure ASL_at(Addr:word); begin Poke(Addr,ASL(Peek(Addr))); end;
Procedure LSR_at(Addr:word); begin Poke(Addr,LSR(Peek(Addr))); end;
Procedure DEC_at(Addr:word); begin Poke(Addr,SetNZ(Peek(Addr)-1)); end;
Procedure INC_at(Addr:word); begin Poke(Addr,SetNZ(Peek(Addr)+1)); end;

Procedure BIT_at(Addr:byte);
 Var
  M:Byte;
 begin
  M:=Peek(Addr);
  SetFlags([ZFg],(A and M)=0);
  SetFlags([VFg],s6 in sw(M));
  SetFlags([NFg],s7 in sw(M));
 end;

Type
 TOpCode=function:byte;

{$F+}
Function op00:byte; { BRK      } begin op00:=7; BRK end;
Function op01:byte; { ORA IndX } begin op01:=6; A:=SetNZ(A or Peek(IndX)) end;

Function op05:byte; { ORA ZPge } begin op05:=3; A:=SetNZ(A or PeekZPge) end;
Function op06:byte; { ASL ZPge } begin op06:=5; ASL_at(Imm) end;

Function op08:byte; { PHP      } begin op08:=3; push(byte(P)); end;
Function op09:byte; { ORA Imm  } begin op09:=2; A:=SetNZ(A or Imm); end;
Function op0A:byte; { ASL      } begin op0A:=2; A:=ASL(A); end;

function op0D:byte; { ORA Abso } begin op0D:=4; A:=SetNZ(A or Peek(Abso)); end;
function op0E:byte; { ASL Abso } begin op0E:=6; ASL_at(Abso); end;

function op10:byte; { BPL Rela } begin op10:=2; Branch(Not (NFg in P)); end;
function op11:byte; { ORA IndY } begin op11:=5; A:=SetNZ(A or Peek(IndY)); end;

function op15:byte; { ORA ZPgX } begin op15:=4; A:=SetNZ(A or Peek(ZPgX)); end;
function op16:byte; { ASL ZPgX } begin op16:=6; ASL_at(ZPgX); end;

function op18:byte; { CLC      } begin op18:=2; P:=P-[CFg];{Exclude(P,CFg);} end;
function op19:byte; { ORA AbsY } begin op19:=4; A:=SetNZ(A or Peek(AbsY)); end;

function op1D:byte; { ORA AbsX } begin op1D:=4; A:=SetNZ(A or Peek(AbsX)); end;
function op1E:byte; { ASL AbsX } begin op1E:=7; ASL_at(AbsX); end;

function op20:byte; { JSR Abso } begin op20:=6; PushW(PC+1); PC:=Abso; end;
function op21:byte; { AND IndX } begin op21:=6; A:=SetNZ(A and Peek(IndX)); end;

function op24:byte; { BIT ZPge } begin op24:=3; BIT_at(ZPge); end;
function op25:byte; { AND ZPge } begin op25:=3; A:=SetNZ(A and PeekZPge); end;
function op26:byte; { ROL ZPge } begin op26:=5; ROL_at(Imm); end;

function op28:byte; { PLP      } begin op28:=4; byte(P):=pop; end;
function op29:byte; { AND Imm  } begin op29:=2; A:=SetNZ(A and Imm); end;
function op2A:byte; { ROL      } begin op2A:=2; A:=ROL(A); end;

function op2C:byte; { BIT Abso } begin op2C:=4; BIT_at(Abso); end;
function op2D:byte; { AND Abso } begin op2D:=4; A:=SetNZ(A and Peek(Abso)); end;
function op2E:byte; { ROL Abso } begin op2E:=6; ROL_at(Abso); end;

function op30:byte; { BMI Rela } begin op30:=2; Branch(NFg in P); end;
function op31:byte; { AND IndY } begin op31:=5; A:=SetNZ(A and Peek(IndY)); end;

function op35:byte; { AND ZPgX } begin op35:=4; A:=SetNZ(A and Peek(ZPgX)); end;
function op36:byte; { ROL ZPgX } begin op36:=6; ROL_at(ZPgX); end;

function op38:byte; { SEC      } begin op38:=2; P:=P+[CFg]{Include(P,CFg)}; end;
function op39:byte; { AND AbsY } begin op39:=4; A:=SetNZ(A and Peek(AbsY)); end;

function op3D:byte; { AND AbsX } begin op3D:=4; A:=SetNZ(A and Peek(AbsX)); end;
function op3E:byte; { ROL AbsX } begin op3E:=7; ROL_at(AbsX); end;

function op40:byte; { RTI      } begin op40:=6; byte(P):=pop; PC:=(pop shl 8 +pop); end;
function op41:byte; { EOR IndX } begin op41:=6; A:=SetNZ(A xor Peek(IndX)); end;

function op45:byte; { EOR ZPge } begin op45:=3; A:=SetNZ(A xor PeekZPge); end;
function op46:byte; { LSR ZPge } begin op46:=5; LSR_at(Imm); end;

function op48:byte; { PHA      } begin op48:=3; Push(A); end;
function op49:byte; { EOR Imm  } begin op49:=2; A:=SetNZ(A xor Imm); end;
function op4A:byte; { LSR      } begin op4A:=2; A:=LSR(A); end;

function op4C:byte; { JMP Abso } begin op4C:=3; PC:=Abso; end;
function op4D:byte; { EOR Abso } begin op4D:=4; A:=SetNZ(A xor Peek(Abso)); end;
function op4E:byte; { LSR Abso } begin op4E:=6; LSR_at(Abso); end;

function op50:byte; { BVC      } begin op50:=2; Branch(Not(VFg in P)); end;
function op51:byte; { EOR IndY } begin op51:=5; A:=SetNZ(A xor Peek(IndY)); end;

function op55:byte; { EOR ZPgX } begin op55:=4; A:=SetNZ(A xor Peek(ZPgX)); end;
function op56:byte; { LSR ZPgX } begin op56:=6; LSR_at(ZPgX); end;

function op58:byte; { CLI      } begin op58:=2; P:=P-[IFg];{Exclude(P,IFg);} end;
function op59:byte; { EOR AbsY } begin op59:=4; A:=SetNZ(A xor Peek(AbsY)); end;

function op5D:byte; { EOR AbsX } begin op5D:=4; A:=SetNZ(A xor Peek(AbsX)); end;
function op5E:byte; { LSR AbsX } begin op5E:=7; LSR_at(AbsX); end;

function op60:byte; { RTS      } begin op60:=6; PC:=(pop shl 8 +pop)+1; end;
function op61:byte; { ADC IndX } begin op61:=6; ADC(Peek(IndX)); end;

function op65:byte; { ADC ZPge } begin op65:=3; ADC(PeekZPge); end;
function op66:byte; { ROR ZPge } begin op66:=5; ROR_at(ZPge); end;

function op68:byte; { PLA      } begin op68:=4; A:=SetNZ(POP); end;
function op69:byte; { ADC Imm  } begin op69:=2; ADC(Imm); end;
function op6A:byte; { ROR      } begin op6A:=2; A:=ROR(A); end;

function op6C:byte; { JMP Ind  } begin op6C:=5; PC:=Ind; end;
function op6D:byte; { ADC Abso } begin op6D:=4; ADC(Peek(Abso)); end;
function op6E:byte; { ROR Abso } begin op6E:=6; ROR_at(Abso); end;

function op70:byte; { BVS      } begin op70:=2; Branch(VFg in P); end;
function op71:byte; { ADC IndY } begin op71:=5; ADC(Peek(IndY)); end;

function op75:byte; { ADC ZPgX } begin op75:=4; ADC(Peek(ZPgX)); end;
function op76:byte; { ROR ZPgX } begin op76:=6; ROR_at(ZPgX); end;

function op78:byte; { SEI      } begin op78:=2; P:=P+[IFg]{Include(P,Ifg)}; end;
function op79:byte; { ADC AbsY } begin op79:=4; ADC(Peek(AbsY)); end;

function op7D:byte; { ADC AbsX } begin op7D:=4; ADC(Peek(AbsX)); end;
function op7E:byte; { ROR AbsX } begin op7E:=7; ROR_at(AbsX); end;

function op81:byte; { STA IndX } begin op81:=6; Poke(IndX,A); end;

function op84:byte; { STY ZPge } begin op84:=3; Poke(ZPge,Y); end;
function op85:byte; { STA ZPge } begin op85:=3; Poke(ZPge,A); end;
function op86:byte; { STX ZPge } begin op86:=3; Poke(ZPge,X); end;

function op88:byte; { DEY      } begin op88:=2; Y:=SetNZ(Y-1); end;
function op8A:byte; { TXA      } begin op8A:=2; A:=SetNZ(X); end;

function op8C:byte; { STY Abso } begin op8C:=4; Poke(Abso,Y); end;
function op8D:byte; { STA Abso } begin op8D:=4; Poke(Abso,A); end;
function op8E:byte; { STX Abso } begin op8E:=4; Poke(Abso,X); end;

function op90:byte; { BCC Rela } begin op90:=2; Branch(Not (Cfg in P)); end;
function op91:byte; { STA IndY } begin op91:=5; Poke(IndY,A); end;

function op94:byte; { STY ZPgX } begin op94:=4; Poke(ZPgX,Y); end;
function op95:byte; { STA ZPgX } begin op95:=4; Poke(ZPgX,A); end;
function op96:byte; { STX ZPgX } begin op96:=4; Poke(ZPgX,X); end;

function op98:byte; { TYA      } begin op98:=2; A:=SetNZ(Y); end;
function op99:byte; { STA AbsY } begin op99:=4; Poke(AbsY,A); end;
function op9A:byte; { TXS      } begin op9A:=2; S:=X; end;

function op9D:byte; { STA AbsX } begin op9D:=4; Poke(AbsX,A); end;

function opA0:byte; { LDY Imm  } begin opA0:=2; Y:=SetNZ(Imm); end;
function opA1:byte; { LDA IndX } begin opA1:=6; A:=PeekNZ(IndX); end;
function opA2:byte; { LDX Imm  } begin opA2:=2; X:=SetNZ(Imm); end;

function opA4:byte; { LDY ZPge } begin opA4:=3; Y:=PeekNZZPge; end;
function opA5:byte; { LDA ZPge } begin opA5:=3; A:=PeekNZZPge; end;
function opA6:byte; { LDX ZPge } begin opA6:=3; X:=PeekNZZPge; end;

function opA8:byte; { TAY      } begin opA8:=2; Y:=SetNZ(A); end;
function opA9:byte; { LDA Imm  } begin opA9:=2; A:=SetNZ(Imm); end;
function opAA:byte; { TAX      } begin opAA:=2; X:=SetNZ(A); end;

function opAC:byte; { LDY Abso } begin opAC:=4; Y:=PeekNZ(Abso); end;
function opAD:byte; { LDA Abso } begin opAD:=4; A:=PeekNZ(Abso); end;
function opAE:byte; { LDX Abso } begin opAE:=4; X:=PeekNZ(Abso); end;

function opB0:byte; { BCS Rela } begin opB0:=2; Branch(CFg in P); end;
function opB1:byte; { LDA IndY } begin opB1:=5; A:=PeekNZ(IndY); end;

function opB4:byte; { LDY ZPgX } begin opB4:=4; Y:=PeekNZ(ZPgX); end;
function opB5:byte; { LDA ZPgX } begin opB5:=4; A:=PeekNZ(ZPgX); end;
function opB6:byte; { LDX ZPgY } begin opB6:=4; X:=PeekNZ(ZPgY); end;

function opB8:byte; { CLV      } begin opB8:=2; P:=P-[VFg];{Exclude(P,VFg);} end;
function opB9:byte; { LDA AbsY } begin opB9:=4; A:=PeekNZ(AbsY); end;
function opBA:byte; { TSX      } begin opBA:=2; X:=SetNZ(S); end;

function opBC:byte; { LDY AbsX } begin opBC:=4; Y:=PeekNZ(AbsX); end;
function opBD:byte; { LDA AbsX } begin opBD:=4; A:=PeekNZ(AbsX); end;
function opBE:byte; { LDX AbsY } begin opBE:=4; X:=PeekNZ(AbsY); end;

function opC0:byte; { CPY Imm  } begin opC0:=2; {NZC=}CMP(Y,Imm); end;
function opC1:byte; { CMP IndX } begin opC1:=6; {NZC=}CMP(A,Peek(IndX)); end;

function opC4:byte; { CPY ZPge } begin opC4:=3; {NZC=}CMP(Y,PeekZPge); end;
function opC5:byte; { CMP ZPge } begin opC5:=3; {NZC=}CMP(A,PeekZPge); end;
function opC6:byte; { DEC ZPge } begin opC6:=5; DEC_at(ZPge); end;

function opC8:byte; { INY      } begin opC8:=2; Y:=SetNZ(Y+1); end;
function opC9:byte; { CMP Imm  } begin opC9:=2; {NZC=}CMP(A,Imm); end;
function opCA:byte; { DEX      } begin opCA:=2; X:=SetNZ(X-1); end;

function opCC:byte; { CPY Abso } begin opCC:=4; {NZC=}CMP(Y,Peek(Abso)); end;
function opCD:byte; { CMP Abso } begin opCD:=4; {NZC=}CMP(A,Peek(Abso)); end;
function opCE:byte; { DEC Abso } begin opCE:=6; DEC_at(Abso); end;

function opD0:byte; { BNE Rela } begin opD0:=2; Branch(not (Zfg in P)); end;
function opD1:byte; { CMP IndY } begin opD1:=5; {NZC=}CMP(A,Peek(IndY)); end;

function opD5:byte; { CMP ZPgX } begin opD5:=4; {NZC=}CMP(A,Peek(ZPgX)); end;
function opD6:byte; { DEC ZPgX } begin opD6:=6; DEC_at(ZPgX); end;

function opD8:byte; { CLD      } begin opD8:=2; P:=P-[CFg];{Exclude(P,CFg);} end;
function opD9:byte; { CMP AbsY } begin opD9:=4; {NZC=}CMP(A,Peek(AbsY)); end;

function opDD:byte; { CMP AbsX } begin opDD:=4; {NZC=}CMP(A,Peek(AbsX)); end;
function opDE:byte; { DEC AbsX } begin opDE:=7; DEC_at(AbsX); end;

function opE0:byte; { CPX Imm  } begin opE0:=2; {NZC=}CMP(X,Imm); end;
function opE1:byte; { SBC IndX } begin opE1:=6; SBC(Peek(IndX)); end;

function opE4:byte; { CPX ZPge } begin opE4:=3; {NZC=}CMP(X,PeekZPge); end;
function opE5:byte; { SBC ZPge } begin opE5:=3; SBC(PeekZPge); end;
function opE6:byte; { INC ZPge } begin opE6:=5; INC_at(ZPge); end;

function opE8:byte; { INX      } begin opE8:=2; X:=SetNZ(X+1); end;
function opE9:byte; { SBC Imm  } begin opE9:=2; SBC(Imm); end;
function opEA:byte; { NOP      } begin opEA:=2; { nothing ! }; end;

function opEC:byte; { CPX Abso } begin opEC:=4; {NZC=}CMP(X,Peek(Abso)); end;
function opED:byte; { SBC Abso } begin opED:=4; SBC(Peek(Abso)); end;
function opEE:byte; { INC Abso } begin opEE:=6; INC_at(Abso); end;

function opF0:byte; { BEQ Rela } begin opF0:=2; Branch(ZFg in P); end;
function opF1:byte; { SBC IndY } begin opF1:=5; SBC(Peek(IndY)); end;

function opF5:byte; { SBC ZPgX } begin opF5:=4; SBC(Peek(ZPgX)); end;
function opF6:byte; { INC ZPgX } begin opF6:=6; INC_at(ZPgX); end;

function opF8:byte; { SED      } begin opF8:=2; Halt(hSED); end;
function opF9:byte; { SBC AbsY } begin opF9:=4; SBC(Peek(AbsY)); end;

function opFD:byte; { SBC AbsX } begin opFD:=4; SBC(Peek(AbsX)); end;
function opFE:byte; { INC AbsX } begin opFE:=7; INC_at(AbsX); end;

function oJAM:byte; begin Halt(hJAM) end;

function oIEC:byte;
 begin
  if PC<$E000 then Halt(hJAM);
  Case Imm of
     $00 : begin
            Bank0^[$90]:=Bank0^[$90] or IEC.Output(Bank0^[$95],Bank0^[$A3] and $80);
            P:=P-[Cfg];{Exclude(P,CFg);}
            PC:=$EDAC;
           end;
     $01 : begin
            Bank0^[$90]:=Bank0^[$90] or IEC.OutputATN(Bank0^[$95]);
            P:=P-[Cfg];{Exclude(P,CFg);}
            PC:=$EDAC;
           end;

     $02 : begin
            Bank0^[$90]:=Bank0^[$90] or IEC.OutputSec(Bank0^[$95]);
            P:=P-[Cfg];{Exclude(P,CFg);}
            PC:=$EDAC;
           end;
     $03 : begin
            Bank0^[$90]:=IEC.Input(A);
            SetNZ(A);
            P:=P-[Cfg];{Exclude(P,CFg);}
            PC:=$EDAC;
           end;
     $04 : PC:=$EDFB; { IECSetATN;     Only needed for real IEC }
     $05 : PC:=$EDAC; { IECRelATN;     Only needed for real IEC }
     $06 : PC:=$EDAC; { IECTurnaround; Only needed for real IEC }
     $07 : PC:=$EDAC; { IECRelease;    Only needed for real IEC }
     else  Halt(hIEC);
  end;
 end;
{$F-}

Const
 OpTable:array[0..255] of TOpCode=(
  op00,op01,oJAM,oJAM,oJAM,op05,op06,oJAM,op08,op09,op0A,oJAM,oJAM,op0D,op0E,oJAM,
  op10,op11,oJAM,oJAM,oJAM,op15,op16,oJAM,op18,op19,oJAM,oJAM,oJAM,op1D,op1E,oJAM,
  op20,op21,oJAM,oJAM,op24,op25,op26,oJAM,op28,op29,op2A,oJAM,op2C,op2D,op2E,oJAM,
  op30,op31,oJAM,oJAM,oJAM,op35,op36,oJAM,op38,op39,oJAM,oJAM,oJAM,op3D,op3E,oJAM,
  op40,op41,oJAM,oJAM,oJAM,op45,op46,oJAM,op48,op49,op4A,oJAM,op4C,op4D,op4E,oJAM,
  op50,op51,oJAM,oJAM,oJAM,op55,op56,oJAM,op58,op59,oJAM,oJAM,oJAM,op5D,op5E,oJAM,
  op60,op61,oJAM,oJAM,oJAM,op65,op66,oJAM,op68,op69,op6A,oJAM,op6C,op6D,op6E,oJAM,
  op70,op71,oJAM,oJAM,oJAM,op75,op76,oJAM,op78,op79,oJAM,oJAM,oJAM,op7D,op7E,oJAM,
  oJAM,op81,oJAM,oJAM,op84,op85,op86,oJAM,op88,oJAM,op8A,oJAM,op8C,op8D,op8E,oJAM,
  op90,op91,oJAM,oJAM,op94,op95,op96,oJAM,op98,op99,op9A,oJAM,oJAM,op9D,oJAM,oJAM,
  opA0,opA1,opA2,oJAM,opA4,opA5,opA6,oJAM,opA8,opA9,opAA,oJAM,opAC,opAD,opAE,oJAM,
  opB0,opB1,oJAM,oJAM,opB4,opB5,opB6,oJAM,opB8,opB9,opBA,oJAM,opBC,opBD,opBE,oJAM,
  opC0,opC1,oJAM,oJAM,opC4,opC5,opC6,oJAM,opC8,opC9,opCA,oJAM,opCC,opCD,opCE,oJAM,
  opD0,opD1,oJAM,oJAM,oJAM,opD5,opD6,oJAM,opD8,opD9,oJAM,oJAM,oJAM,opDD,opDE,oJAM,
  opE0,opE1,oJAM,oJAM,opE4,opE5,opE6,oJAM,opE8,opE9,opEA,oJAM,opEC,opED,opEE,oJAM,
  opF0,opF1,oIEC,oJAM,oJAM,opF5,opF6,oJAM,opF8,opF9,oJAM,oJAM,oJAM,opFD,opFE,oJAM
 );

Procedure Run;
 begin
  while Cycles>0 do begin
   IR:=Peek(PC);
   Inc(PC);
   Dec(Cycles,OpTable[IR and 255]);
  end;
 end;


Const
 OpCycles:array[0..255] of byte=(
   { 0 1 2 3 4 5 6 7 8 9 A B C D E F 0 1 2 3 4 5 6 7 8 9 A B C D E F}
{00} 7,6,0,0,0,3,5,0,3,2,2,0,0,4,6,0,2,5,0,0,0,4,6,0,2,4,0,0,0,4,7,0,
{20} 6,6,0,0,3,3,5,0,4,2,2,0,4,4,6,0,2,5,0,0,0,4,6,0,2,4,0,0,0,4,7,0,
{40} 6,6,0,0,0,3,5,0,3,2,2,0,3,4,6,0,2,5,0,0,0,4,6,0,2,4,0,0,0,4,7,0,
{60} 6,6,0,0,0,3,5,0,4,2,2,0,5,4,6,0,2,5,0,0,0,4,6,0,2,4,0,0,0,4,7,0,
{80} 0,6,0,0,3,3,3,0,2,0,2,0,4,4,4,0,2,5,0,0,4,4,4,0,2,4,2,0,0,4,0,0,
{A0} 2,6,2,0,3,3,3,0,2,2,2,0,4,4,4,0,2,5,0,0,4,4,4,0,2,4,2,0,4,4,4,0,
{C0} 2,6,0,0,3,3,5,0,2,2,2,0,4,4,6,0,2,5,0,0,0,4,6,0,2,4,0,0,0,4,7,0,
{E0} 2,6,0,0,3,3,5,0,2,2,2,0,4,4,6,0,2,5,0,0,0,4,6,0,2,4,0,0,0,4,7,0
 );

Procedure RunCase;
begin
 While Cycles>0 do begin
  Fetch; { Memory.Fetch or overrided by Debug.Fetch }
  case IR of
   $00 { BRK      } : BRK;
   $01 { ORA IndX } : A:=SetNZ(A or Peek(IndX));
   $05 { ORA ZPge } : A:=SetNZ(A or PeekZPge);
   $06 { ASL ZPge } : ASL_at(Imm);
   $08 { PHP      } : push(byte(P));
   $09 { ORA Imm  } : A:=SetNZ(A or Imm);
   $0A { ASL      } : A:=ASL(A);
   $0D { ORA Abso } : A:=SetNZ(A or Peek(Abso));
   $0E { ASL Abso } : ASL_at(Abso);
   $10 { BPL Rela } : Branch(Not (NFg in P));
   $11 { ORA IndY } : A:=SetNZ(A or Peek(IndY));
   $15 { ORA ZPgX } : A:=SetNZ(A or Peek(ZPgX));
   $16 { ASL ZPgX } : ASL_at(ZPgX);
   $18 { CLC      } : P:=P-[Cfg];{Exclude(P,CFg);}
   $19 { ORA AbsY } : A:=SetNZ(A or Peek(AbsY));
   $1D { ORA AbsX } : A:=SetNZ(A or Peek(AbsX));
   $1E { ASL AbsX } : ASL_at(AbsX);
   $20 { JSR Abso } : begin PushW(PC+1); PC:=Abso; end;
   $21 { AND IndX } : A:=SetNZ(A and Peek(IndX));
   $24 { BIT ZPge } : BIT_at(ZPge);
   $25 { AND ZPge } : A:=SetNZ(A and PeekZPge);
   $26 { ROL ZPge } : ROL_at(Imm);
   $28 { PLP      } : byte(P):=pop;
   $29 { AND Imm  } : A:=SetNZ(A and Imm);
   $2A { ROL      } : A:=ROL(A);
   $2C { BIT Abso } : BIT_at(Abso);
   $2D { AND Abso } : A:=SetNZ(A and Peek(Abso));
   $2E { ROL Abso } : ROL_at(Abso);
   $30 { BMI Rela } : Branch(NFg in P);
   $31 { AND IndY } : A:=SetNZ(A and Peek(IndY));
   $35 { AND ZPgX } : A:=SetNZ(A and Peek(ZPgX));
   $36 { ROL ZPgX } : ROL_at(ZPgX);
   $38 { SEC      } : P:=P+[CFg];{Include(P,CFg)}
   $39 { AND AbsY } : A:=SetNZ(A and Peek(AbsY));
   $3D { AND AbsX } : A:=SetNZ(A and Peek(AbsX));
   $3E { ROL AbsX } : ROL_at(AbsX);
   $40 { RTI      } : begin byte(P):=pop; PC:=(pop shl 8 +pop); end;
   $41 { EOR IndX } : A:=SetNZ(A xor Peek(IndX));
   $45 { EOR ZPge } : A:=SetNZ(A xor PeekZPge);
   $46 { LSR ZPge } : LSR_at(Imm);
   $48 { PHA      } : Push(A);
   $49 { EOR Imm  } : A:=SetNZ(A xor Imm);
   $4A { LSR      } : A:=LSR(A);
   $4C { JMP Abso } : PC:=Abso;
   $4D { EOR Abso } : A:=SetNZ(A xor Peek(Abso));
   $4E { LSR Abso } : LSR_at(Abso);
   $50 { BVC      } : Branch(Not(VFg in P));
   $51 { EOR IndY } : A:=SetNZ(A xor Peek(IndY));
   $55 { EOR ZPgX } : A:=SetNZ(A xor Peek(ZPgX));
   $56 { LSR ZPgX } : LSR_at(ZPgX);
   $58 { CLI      } : Exclude(P,IFg);
   $59 { EOR AbsY } : A:=SetNZ(A xor Peek(AbsY));
   $5D { EOR AbsX } : A:=SetNZ(A xor Peek(AbsX));
   $5E { LSR AbsX } : LSR_at(AbsX);
   $60 { RTS      } : PC:=(pop shl 8 +pop)+1;
   $61 { ADC IndX } : ADC(Peek(IndX));
   $65 { ADC ZPge } : ADC(PeekZPge);
   $66 { ROR ZPge } : ROR_at(ZPge);
   $68 { PLA      } : A:=SetNZ(POP);
   $69 { ADC Imm  } : ADC(Imm);
   $6A { ROR      } : A:=ROR(A);
   $6C { JMP Ind  } : PC:=Ind;
   $6D { ADC Abso } : ADC(Peek(Abso));
   $6E { ROR Abso } : ROR_at(Abso);
   $70 { BVS      } : Branch(VFg in P);
   $71 { ADC IndY } : ADC(Peek(IndY));
   $75 { ADC ZPgX } : ADC(Peek(ZPgX));
   $76 { ROR ZPgX } : ROR_at(ZPgX);
   $78 { SEI      } : P:=P+[IFg];{Include(P,Ifg);}
   $79 { ADC AbsY } : ADC(Peek(AbsY));
   $7D { ADC AbsX } : ADC(Peek(AbsX));
   $7E { ROR AbsX } : ROR_at(AbsX);
   $81 { STA IndX } : Poke(IndX,A);
   $84 { STY ZPge } : Poke(ZPge,Y);
   $85 { STA ZPge } : Poke(ZPge,A);
   $86 { STX ZPge } : Poke(ZPge,X);
   $88 { DEY      } : Y:=SetNZ(Y-1);
   $8A { TXA      } : A:=SetNZ(X);
   $8C { STY Abso } : Poke(Abso,Y);
   $8D { STA Abso } : Poke(Abso,A);
   $8E { STX Abso } : Poke(Abso,X);
   $90 { BCC Rela } : Branch(Not (Cfg in P));
   $91 { STA IndY } : Poke(IndY,A);
   $94 { STY ZPgX } : Poke(ZPgX,Y);
   $95 { STA ZPgX } : Poke(ZPgX,A);
   $96 { STX ZPgX } : Poke(ZPgX,X);
   $98 { TYA      } : A:=SetNZ(Y);
   $99 { STA AbsY } : Poke(AbsY,A);
   $9A { TXS      } : S:=X;
   $9D { STA AbsX } : Poke(AbsX,A);
   $A0 { LDY Imm  } : Y:=SetNZ(Imm);
   $A1 { LDA IndX } : A:=PeekNZ(IndX);
   $A2 { LDX Imm  } : X:=SetNZ(Imm);
   $A4 { LDY ZPge } : Y:=PeekNZZPge;
   $A5 { LDA ZPge } : A:=PeekNZZPge;
   $A6 { LDX ZPge } : X:=PeekNZZPge;
   $A8 { TAY      } : Y:=SetNZ(A);
   $A9 { LDA Imm  } : A:=SetNZ(Imm);
   $AA { TAX      } : X:=SetNZ(A);
   $AC { LDY Abso } : Y:=PeekNZ(Abso);
   $AD { LDA Abso } : A:=PeekNZ(Abso);
   $AE { LDX Abso } : X:=PeekNZ(Abso);
   $B0 { BCS Rela } : Branch(CFg in P);
   $B1 { LDA IndY } : A:=PeekNZ(IndY);
   $B4 { LDY ZPgX } : Y:=PeekNZ(ZPgX);
   $B5 { LDA ZPgX } : A:=PeekNZ(ZPgX);
   $B6 { LDX ZPgY } : X:=PeekNZ(ZPgY);
   $B8 { CLV      } : Exclude(P,VFg);
   $B9 { LDA AbsY } : A:=PeekNZ(AbsY);
   $BA { TSX      } : X:=SetNZ(S);
   $BC { LDY AbsX } : Y:=PeekNZ(AbsX);
   $BD { LDA AbsX } : A:=PeekNZ(AbsX);
   $BE { LDX AbsY } : X:=PeekNZ(AbsY);
   $C0 { CPY Imm  } : {NZC=}CMP(Y,Imm);
   $C1 { CMP IndX } : {NZC=}CMP(A,Peek(IndX));
   $C4 { CPY ZPge } : {NZC=}CMP(Y,PeekZPge);
   $C5 { CMP ZPge } : {NZC=}CMP(A,PeekZPge);
   $C6 { DEC ZPge } : DEC_at(ZPge);
   $C8 { INY      } : Y:=SetNZ(Y+1);
   $C9 { CMP Imm  } : {NZC=}CMP(A,Imm);
   $CA { DEX      } : X:=SetNZ(X-1);
   $CC { CPY Abso } : {NZC=}CMP(Y,Peek(Abso));
   $CD { CMP Abso } : {NZC=}CMP(A,Peek(Abso));
   $CE { DEC Abso } : DEC_at(Abso);
   $D0 { BNE Rela } : Branch(not (Zfg in P));
   $D1 { CMP IndY } : {NZC=}CMP(A,Peek(IndY));
   $D5 { CMP ZPgX } : {NZC=}CMP(A,Peek(ZPgX));
   $D6 { DEC ZPgX } : DEC_at(ZPgX);
   $D8 { CLD      } : Exclude(P,CFg);
   $D9 { CMP AbsY } : {NZC=}CMP(A,Peek(AbsY));
   $DD { CMP AbsX } : {NZC=}CMP(A,Peek(AbsX));
   $DE { DEC AbsX } : DEC_at(AbsX);
   $E0 { CPX Imm  } : {NZC=}CMP(X,Imm);
   $E1 { SBC IndX } : SBC(Peek(IndX));
   $E4 { CPX ZPge } : {NZC=}CMP(X,PeekZPge);
   $E5 { SBC ZPge } : SBC(PeekZPge);
   $E6 { INC ZPge } : INC_at(ZPge);
   $E8 { INX      } : X:=SetNZ(X+1);
   $E9 { SBC Imm  } : SBC(Imm);
   $EA { NOP      } : { nothing ! };
   $EC { CPX Abso } : {NZC=}CMP(X,Peek(Abso));
   $ED { SBC Abso } : SBC(Peek(Abso));
   $EE { INC Abso } : INC_at(Abso);
   $F0 { BEQ Rela } : Branch(ZFg in P);
   $F1 { SBC IndY } : SBC(Peek(IndY));

   $F2 { Extension OpCode } : begin
    if PC<$E000 then Halt(hJAM);
    Case Imm of
     $00 : begin
            Bank0^[$90]:=Bank0^[$90] or IEC.Output(Bank0^[$95],Bank0^[$A3] and $80);
            Exclude(P,CFg);
            PC:=$EDAC;
           end;
     $01 : begin
            Bank0^[$90]:=Bank0^[$90] or IEC.OutputATN(Bank0^[$95]);
            Exclude(P,CFg);
            PC:=$EDAC;
           end;

     $02 : begin
            Bank0^[$90]:=Bank0^[$90] or IEC.OutputSec(Bank0^[$95]);
            Exclude(P,CFg);
            PC:=$EDAC;
           end;
     $03 : begin
            Bank0^[$90]:=IEC.Input(A);
            SetNZ(A);
            Exclude(P,CFg);
            PC:=$EDAC;
           end;
     $04 : PC:=$EDFB; { IECSetATN;     Only needed for real IEC }
     $05 : PC:=$EDAC; { IECRelATN;     Only needed for real IEC }
     $06 : PC:=$EDAC; { IECTurnaround; Only needed for real IEC }
     $07 : PC:=$EDAC; { IECRelease;    Only needed for real IEC }
     else  Halt(hIEC);
    end;
   end;
   $F5 { SBC ZPgX } : SBC(Peek(ZPgX));
   $F6 { INC ZPgX } : INC_at(ZPgX);
   $F8 { SED      } : Halt(hSED);
   $F9 { SBC AbsY } : SBC(Peek(AbsY));
   $FD { SBC AbsX } : SBC(Peek(AbsX));
   $FE { INC AbsX } : INC_at(AbsX);
   else Halt(hJAM);
  end;
  Dec(Cycles,OpCycles[IR]);
 end;
end;

end.