Unit Debug;

Interface

Procedure Fetch;             { Override Memory.Fetch     }
Procedure VIC_Mode;          { Override Video.VIC_Mode   }
Procedure VIC_Update;        { Override Video.VIC_Update }
Function KeyPressed:boolean; { Override Tools.KeyPressed }
Function ReadKey:byte;       { Override Tools.ReadKey    }

Implementation

Uses
 Memory,Video,HaltCode,Tools,Dos;

Var
 Keys:string;         { Keyboard buffer }
 DebugActive:boolean; { Debug Mode }
 BreakPoint:word;     { BreakPoint Address }

Type
 TParm=(
  pImpl, { no parameters }
  pAbso, { word }
  pZPge, { byte }
  pImm,  { #byte }
  pAbsX, { word+x }
  pAbsY, { word+y }
  pIndX, { [byte+x] = array[x]^ }
  pIndY, { [byte]+y = array^[y] }
  pZPgX, { byte+x }
  pZPgY, { byte+y }
  pRela, { shortint(byte) }
  pInd   { [word] }
 );

 TOpCode=record
  Name:array[0..2] of char;
  Parm:TParm;
 end;

Const
 OpCode:Array[0..255] of TOpCode = (
 {00-0F}
  (Name:'BRK'; Parm:pImpl),  (Name:'ORA'; Parm:pIndX),
  (Name:'JAM'; Parm:pImpl),  (Name:'SLO'; Parm:pIndX),
  (Name:'NOP'; Parm:pImpl),  (Name:'ORA'; Parm:pZPge),
  (Name:'ASL'; Parm:pZPge),  (Name:'SLO'; Parm:pZPge),
  (Name:'PHP'; Parm:pImpl),  (Name:'ORA'; Parm:PImm),
  (Name:'ASL'; Parm:pImpl),  (Name:'ANC'; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'ORA'; Parm:pAbso),
  (Name:'ASL'; Parm:pAbso),  (Name:'   '; Parm:pImpl),
 {10-1F}
  (Name:'BPL'; Parm:pRela),  (Name:'ORA'; Parm:pIndY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'ORA'; Parm:pZPgX),
  (Name:'ASL'; Parm:pZPgX),  (Name:'   '; Parm:pImpl),
  (Name:'CLC'; Parm:pImpl),  (Name:'ORA'; Parm:pAbsY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'ORA'; Parm:pAbsX),
  (Name:'ASL'; Parm:pAbsX),  (Name:'   '; Parm:pImpl),
 {20-2F}
  (Name:'JSR'; Parm:pAbso),  (Name:'AND'; Parm:pIndX),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'BIT'; Parm:pZPge),  (Name:'AND'; Parm:pZPge),
  (Name:'ROL'; Parm:pZPge),  (Name:'   '; Parm:pImpl),
  (Name:'PLP'; Parm:pImpl),  (Name:'AND'; Parm:PImm),
  (Name:'ROL'; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'BIT'; Parm:pAbso),  (Name:'AND'; Parm:pAbso),
  (Name:'ROL'; Parm:pAbso),  (Name:'   '; Parm:pImpl),
 {30-3F}
  (Name:'BMI'; Parm:pRela),  (Name:'AND'; Parm:pIndY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'AND'; Parm:pZPgX),
  (Name:'ROL'; Parm:pZPgX),  (Name:'   '; Parm:pImpl),
  (Name:'SEC'; Parm:pImpl),  (Name:'AND'; Parm:pAbsY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'AND'; Parm:pAbsX),
  (Name:'ROL'; Parm:pAbsX),  (Name:'   '; Parm:pImpl),
 {40-4F}
  (Name:'RTI'; Parm:pImpl),  (Name:'EOR'; Parm:pIndX),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'EOR'; Parm:pZPge),
  (Name:'LSR'; Parm:pZPge),  (Name:'   '; Parm:pImpl),
  (Name:'PHA'; Parm:pImpl),  (Name:'EOR'; Parm:pImm ),
  (Name:'LSR'; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'JMP'; Parm:pAbso),  (Name:'EOR'; Parm:pAbso),
  (Name:'LSR'; Parm:pAbso),  (Name:'   '; Parm:pImpl),
 {50-5F}
  (Name:'BVC'; Parm:pRela),  (Name:'EOR'; Parm:pIndY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'EOR'; Parm:pZPgX),
  (Name:'LSR'; Parm:pZPgX),  (Name:'   '; Parm:pImpl),
  (Name:'CLI'; Parm:pImpl),  (Name:'EOR'; Parm:pAbsY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'EOR'; Parm:pAbsX),
  (Name:'LSR'; Parm:pAbsX),  (Name:'   '; Parm:pImpl),
 {60-6F}
  (Name:'RTS'; Parm:pImpl),  (Name:'ADC'; Parm:pIndX),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'ADC'; Parm:pZPge),
  (Name:'ROR'; Parm:pZPge),  (Name:'   '; Parm:pImpl),
  (Name:'PLA'; Parm:pImpl),  (Name:'ADC'; Parm:PImm),
  (Name:'ROR'; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'JMP'; Parm:pInd ),  (Name:'ADC'; Parm:pAbso),
  (Name:'ROR'; Parm:pAbso),  (Name:'   '; Parm:pImpl),
 {70-7F}
  (Name:'BVS'; Parm:pRela),  (Name:'ADC'; Parm:pIndY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'ADC'; Parm:pZPgX),
  (Name:'ROR'; Parm:pZPgX),  (Name:'   '; Parm:pImpl),
  (Name:'SEI'; Parm:pImpl),  (Name:'ADC'; Parm:pAbsY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'ADC'; Parm:pAbsX),
  (Name:'ROR'; Parm:pAbsX),  (Name:'   '; Parm:pImpl),
 {80-8F}
  (Name:'   '; Parm:pImpl),  (Name:'STA'; Parm:pIndX),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'STY'; Parm:pZPge),  (Name:'STA'; Parm:pZPge),
  (Name:'STX'; Parm:pZPge),  (Name:'   '; Parm:pImpl),
  (Name:'DEY'; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'TXA'; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'STY'; Parm:pAbso),  (Name:'STA'; Parm:pAbso),
  (Name:'STX'; Parm:pAbso),  (Name:'   '; Parm:pImpl),
 {90-9F}
  (Name:'BCC'; Parm:pRela),  (Name:'STA'; Parm:pIndY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'STY'; Parm:pZPgX),  (Name:'STA'; Parm:pZPgX),
  (Name:'STX'; Parm:pZPgY),  (Name:'   '; Parm:pImpl),
  (Name:'TYA'; Parm:pImpl),  (Name:'STA'; Parm:pAbsY),
  (Name:'TXS'; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'STA'; Parm:pAbsX),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
 {A0-AF}
  (Name:'LDY'; Parm:PImm),   (Name:'LDA'; Parm:pIndX),
  (Name:'LDX'; Parm:PImm),   (Name:'   '; Parm:pImpl),
  (Name:'LDY'; Parm:pZPge),  (Name:'LDA'; Parm:pZPge),
  (Name:'LDX'; Parm:pZPge),  (Name:'   '; Parm:pImpl),
  (Name:'TAY'; Parm:pImpl),  (Name:'LDA'; Parm:PImm),
  (Name:'TAX'; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'LDY'; Parm:pAbso),  (Name:'LDA'; Parm:pAbso),
  (Name:'LDX'; Parm:pAbso),  (Name:'   '; Parm:pImpl),
 {B0-BF}
  (Name:'BCS'; Parm:pRela),  (Name:'LDA'; Parm:pIndY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'LDY'; Parm:pZPgX),  (Name:'LDA'; Parm:pZpgX),
  (Name:'LDX'; Parm:pZPgY),  (Name:'   '; Parm:pImpl),
  (Name:'CLV'; Parm:pImpl),  (Name:'LDA'; Parm:pAbsY),
  (Name:'TSX'; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'LDY'; Parm:pAbsX),  (Name:'LDA'; Parm:pAbsX),
  (Name:'LDX'; Parm:pAbsY),  (Name:'   '; Parm:pImpl),
 {C0-CF}
  (Name:'CPY'; Parm:pImm ),  (Name:'CMP'; Parm:pIndX),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'CPY'; Parm:pZPge),  (Name:'CMP'; Parm:pZPge),
  (Name:'DEC'; Parm:pZPge),  (Name:'   '; Parm:pImpl),
  (Name:'INY'; Parm:pImpl),  (Name:'CMP'; Parm:PImm),
  (Name:'DEX'; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'CPY'; Parm:pAbso),  (Name:'CMP'; Parm:pAbso),
  (Name:'DEC'; Parm:pAbso),  (Name:'   '; Parm:pImpl),
 {D0-DF}
  (Name:'BNE'; Parm:pRela),  (Name:'CMP'; Parm:pIndY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'DEC'; Parm:pZPgX),  (Name:'   '; Parm:pImpl),
  (Name:'CLD'; Parm:pImpl),  (Name:'CMP'; Parm:pAbsY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'CMP'; Parm:pAbsX),
  (Name:'DEC'; Parm:pAbsX),  (Name:'   '; Parm:pImpl),
 {E0-EF}
  (Name:'CPX'; Parm:PImm),   (Name:'SBC'; Parm:pIndX),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'CPX'; Parm:pZPge),  (Name:'SBC'; Parm:pZPge),
  (Name:'INC'; Parm:pZPge),  (Name:'   '; Parm:pImpl),
  (Name:'INX'; Parm:pImpl),  (Name:'SBC'; Parm:PImm),
  (Name:'NOP'; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'CPX'; Parm:pAbso),  (Name:'SBC'; Parm:pAbso),
  (Name:'INC'; Parm:pAbso),  (Name:'   '; Parm:pImpl),
 {F0-FF}
  (Name:'BEQ'; Parm:pRela),  (Name:'SBC'; Parm:pIndY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'SBC'; Parm:pZPgX),
  (Name:'INC'; Parm:pZPgX),  (Name:'   '; Parm:pImpl),
  (Name:'SED'; Parm:pImpl),  (Name:'SBC'; Parm:pAbsY),
  (Name:'   '; Parm:pImpl),  (Name:'   '; Parm:pImpl),
  (Name:'   '; Parm:pImpl),  (Name:'SBC'; Parm:pAbsX),
  (Name:'INC'; Parm:pAbsX),  (Name:'   '; Parm:pImpl)
  );

Const
 FlagName:Array[Cfg..NFg] of char='czidb1vn';

Var
 Value:word;

Function HexToInt(s:string):word;
 const
  hx:array[$0..$F] of char='0123456789ABCDEF';
 var
  result:word;
  i,h:byte;
 begin
  result:=0;
  if length(s)>4 then exit;
  for i:=1 to length(s) do begin
   h:=pos(upcase(s[i]),hx);
   if h=0 then exit;
   dec(h); result:=result*16+h;
  end;
  HexToInt:=result;
 end;

Procedure DumpMemory;
 Var
  sAddr:string;
  wAddr,Addr:word;
  dump:string;
  l,b,p:byte;
 begin
  Write('DUMP at : '); ReadLn(sAddr);
  wAddr:=HexToInt(sAddr);
  for l:=0 to 5 do begin
   Write(HexaW(wAddr));
   dump:=' ''';
   for b:=0 to 7 do begin
    p:=peek(wAddr);
    Write(' ',HexaB(p));
    case p of
     7	: dump:=dump+'.';
     9  : dump:=dump+'';
     10 : dump:=dump+'';
     13 : dump:=dump+'';
     else dump:=dump+chr(p);
    end;
    inc(wAddr);
   end;
   write(' -');
   dump:=dump+'''-''';
   for b:=8 to 15 do begin
    p:=peek(wAddr);
    Write(' ',HexaB(p));
    case p of
     7	: dump:=dump+'.';
     9  : dump:=dump+'';
     10 : dump:=dump+'';
     13 : dump:=dump+'';
     else dump:=dump+chr(p);
    end;
    inc(wAddr);
   end;
   writeln(dump,'''');
  end;
 end;

Function GotoAddr:word;
 var
  saddr:string;
  result:word;
 begin
  Write('Go to : '); ReadLn(sAddr);
  Result:=HexToInt(sAddr);
  WriteLn('Next break point at '+HexaW(Result));
  GotoAddr:=Result;
  DebugActive:=False;
 end;

Procedure QuickHelp;
 begin
  WriteLn('CBM64 Debug Mode');
  WriteLn('H for Help');
 end;

Procedure Help;
 begin
  WriteLn('Help :');
  WriteLn('D = dump (address)');
  WriteLn('F = files (dir)');
  WriteLn('G = go (address)');
  WriteLn('K = send key (readkey)');
  WriteLn('L = load (basic)');
  WriteLn('P = proceed (JSR, BNE...)');
  WriteLn('Q = Quit to DOS');
  WriteLn('R = registers');
  WriteLn('S = save (basic)');
  WriteLn('T = trace');
  WriteLn('V = View screen');
  WriteLn('<Esc> return to C=64');
 end;

Procedure DumpRegisters;
 Const        {1...:...10....:...20....:...30....:...40....:}
  Dump:string='A=?? X=?? Y=?? S=?? F=???????? PC=???? IR=?? ';
 var
  f:TFlag;
  h:string;
 begin
   h:=HexaB(A)+HexaB(X)+HexaB(Y)+HexaB(S)+HexaW(PC-1)+HexaB(IR);
   Move(h[1],Dump[3],2);   { A }
   Move(h[3],Dump[8],2);   { X }
   Move(h[5],Dump[13],2);  { Y }
   Move(h[7],Dump[18],2);  { S }
   Move(h[9],Dump[35],4);  { PC }
   Move(h[13],Dump[43],2); { IR }
   h:='';
   for f:=Cfg to NFg do begin
    if f in P then h:=UpCase(FlagName[f])+h else h:=FlagName[f]+h;
   end;
   Move(h[1],Dump[23],8); { P Flags }
   With OpCode[IR] do begin
    h:=Copy(Name,1,3)+' ';
    value:=PeekW(PC);
    Case Parm of
     pImpl : { nothing };
     pAbso : h:=h+'$'+HexaW(Value);
     pZPge : h:=h+'$'+HexaB(Lo(Value));
     pImm  : h:=h+'#$'+HexaB(Lo(Value));
     pAbsX : h:=h+'$'+HexaW(Value)+'+X';
     pAbsY : h:=h+'$'+HexaW(Value)+'+Y';
     pIndX : h:=h+'($'+HexaB(Lo(Value))+'+X)';
     pIndY : h:=h+'($'+HexaB(Lo(Value))+')+Y';
     pZPgX : h:=h+'$'+HexaB(Lo(Value))+'+X';
     pZPgY : h:=h+'$'+HexaB(Lo(Value))+'+Y';
     pRela : h:=h+'$'+HexaW(PC+1+ShortInt(Lo(Value)));
     pInd  : h:=h+'($'+HexaW(Value)+')';
    end;
    Writeln(Dump,h);
    if name[1]=' ' then Halt;
   end;
 end;

Function Proceed:boolean;
 begin
  case IR of
   $20{JSR} : begin BreakPoint:=PC+3; Proceed:=True; end;
   $10{BPL},
   $30{BMI},
   $50{BVC},
   $70{BVS},
   $90{BCC},
   $B0{BCS},
   $D0{BNE},
   $F0{BEQ} : If ShortInt(Lo(Value))>0 then
               Proceed:=False
              else begin
               Proceed:=True;
               BreakPoint:=PC+2;
              end;
   else
    Proceed:=False;
   end;
 end;

Procedure DebugKey;
 Var
  Key:byte;
 begin
  Poke($C6,0);
  Write('Key to send : ');
  Key:=Tools.ReadKey;
  if key=27 then
   WriteLn('Cancel')
  else begin
   Write(HexaB(Key),'-');
   PutKey(Key);
   WriteLn(HexaB(Peek($277)));
  end;
 end;

Type
 TRegInfo=record
  iA,iX,iY,iS,iIR:byte;
  iP:Flags;
  iPC:word;
 end;

Procedure Load;
 Var
  FileName:string;
  f:file;
  RegSaved:TRegInfo;
 begin
  Write('Load from file : '); ReadLn(FileName);
  if pos('.',FileName)=0 then FileName:=FileName+'.B64';
  Assign(f,FileName);
  Reset(f,1);
  If IoResult<>0 then
   WriteLn('File not found')
  else begin
   BlockRead(f,RegSaved,SizeOf(TRegInfo));
   BlockRead(f,BANK0^,SizeOf(TBANK0));
   With RegSaved do begin
    A:=iA; X:=iX; Y:=iY; S:=iS; IR:=iIR;
    P:=iP;
    PC:=iPC;
   end;
   DumpRegisters;
   Close(f);
   WriteLn('File loaded');
  end;
 end;

Procedure Save;
 Var
  FileName:string;
  f:file;
  SaveRegs:TRegInfo;
 begin
  Write('Save to file : '); ReadLn(FileName);
  if pos('.',FileName)=0 then FileName:=FileName+'.B64';
  Assign(f,FileName);
  Rewrite(f,1);
  if IoResult<>0 then
   WriteLn('Unabled to create file')
  else begin
   With SaveRegs do begin
    iA:=A; iX:=X; iY:=Y; iS:=S; iIR:=IR;
    iP:=P;
    iPC:=PC;
   end;
   BlockWrite(f,SaveRegs,SizeOf(TRegInfo));
   BlockWrite(f,BANK0^,SizeOf(TBANK0));
   Close(f);
   WriteLn('File saved');
  end;
 end;

Procedure Files;
 Var
  ff:SearchRec;
 begin
  FindFirst('*.B64',AnyFile,ff);
  While DosError=0 do begin
   WriteLn(ff.Name);
   FindNext(ff);
  end;
 end;

Procedure DebugMode;
 Var
  Loop:boolean;
  i:integer;
 begin
  Loop:=True;
  DebugActive:=False;
  if TextMode then QuickHelp;
  DumpRegisters;
  Repeat
   Case chr(Tools.Readkey) of
     'd','D': DumpMemory;
     'f','F': Files;
     'g','G': begin BreakPoint:=GotoAddr+1; Loop:=False; end;
     'h','H': Help;
     'k','K': DebugKey;
     'l','L': Load;
     'p','P': begin DebugActive:=not Proceed; Loop:=False; end;
     'q','Q': Halt(hESC); { Exit CBM64 }
     'r','R': DumpRegisters;
     's','S': Save;
     't','T': begin DebugActive:=True; Loop:=False; end;
     'v','V': begin
      Video.VIC_Mode;
      For i:=0 to 40*25 do Video.VIC_Update;
      readkey;
      Video.TextMode;
      QuickHelp;
      DumpRegisters;
     end;
     #27 : Loop:=False;
     else loop:=true;
    end;
   Until loop=false;
   if Not DebugActive then Video.VIC_Mode;
 end;

Procedure Fetch;
 begin
  Memory.Fetch;
  if PC=BreakPoint then begin
   WriteLn('*BREAK*');
   BreakPoint:=0;
   DebugActive:=True;
  end;
  if DebugActive then DebugMode;
 end;

Procedure VIC_Mode;
 begin
  Video.VIC_Mode;
  Keys:='';
  DebugActive:=ParamCount>0;
  BreakPoint:=0;
 end;

Procedure VIC_Update;
 begin
  { no video output on debug mode}
  if Not DebugActive then Video.VIC_Update;
 end;

Function KeyPressed:boolean;
 Var
  Key:byte;
 begin
  if (Length(Keys)<255)and(Tools.KeyPressed) then begin
   Key:=Tools.Readkey;
   if Key=27 then DebugActive:=True else Keys:=Keys+Chr(Key);
  end;
  KeyPressed:=Length(Keys)>0;
 end;

Function ReadKey:byte;
 begin
  repeat until KeyPressed;
  ReadKey:=Ord(Keys[1]);
  Keys:=Copy(Keys,2,Length(Keys));
 end;

end.
